@@ -1,5 +1,5 @@
 //
-// Copyright (c) 2000,2001 Brian McNamara and Yannis Smaragdakis
+// Copyright (c) 2000,2001,2002 Brian McNamara and Yannis Smaragdakis
 //
 // Permission to use, copy, modify, distribute and sell this software
 // and its documentation for any purpose is granted without fee,
@@ -30,9 +30,10 @@ namespace fcpp {
 
 // Important to implementation of CurryableN classes
 struct AutoCurryType {};
-namespace {
-AutoCurryType _;   // this is a legal identifier as fcpp::_
-}
+FCPP_MAYBE_NAMESPACE_OPEN
+FCPP_MAYBE_EXTERN AutoCurryType _;   // this is a legal identifier as fcpp::_
+FCPP_MAYBE_NAMESPACE_CLOSE
+
 
 //////////////////////////////////////////////////////////////////////
 // Const
@@ -40,12 +41,11 @@ AutoCurryType _;   // this is a legal identifier as fcpp::_
 
 template <class T>
 struct ConstHelper : public CFunType<T> {
-   T x;
+   const T x;
 public:
    ConstHelper( const T& a ) : x(a) {}
    T operator()() const { return x; }
 };
-
 struct Const {
    template <class T>
    struct Sig : public FunType<T,ConstHelper<T> > {};
@@ -55,6 +55,9 @@ struct Const {
       return ConstHelper<T>(x);
    }
 };
+FCPP_MAYBE_NAMESPACE_OPEN
+FCPP_MAYBE_EXTERN Const const_;   // C++ keyword, so add trailing underscore
+FCPP_MAYBE_NAMESPACE_CLOSE
 
 //////////////////////////////////////////////////////////////////////
 // Binders (through "...of2")
@@ -63,8 +66,8 @@ struct Const {
 template <class Unary, class Arg>
 class binder1of1 
 : public CFunType<typename RT<Unary,Arg>::ResultType> {
-   Unary f;
-   Arg a;
+   const Unary f;
+   const Arg a;
 public:
    binder1of1( const Unary& x, const Arg& y ) : f(x), a(y) {}
    typename RT<Unary,Arg>::ResultType operator()() const { return f(a); }
@@ -80,15 +83,15 @@ inline
       return binder1of1<Unary,Arg>(f,a);
    }
 };
-namespace {
-Bind1of1 bind1of1;
+FCPP_MAYBE_NAMESPACE_OPEN
+FCPP_MAYBE_EXTERN Bind1of1 bind1of1;
 //Curryable2<Bind1of1> bind1of1(_bind1of1);
-}
+FCPP_MAYBE_NAMESPACE_CLOSE
 
 template <class Binary, class Arg1>
 class binder1of2 {
-   Binary f;
-   Arg1 x;
+   const Binary f;
+   const Arg1 x;
 public:
    binder1of2( const Binary& a, const Arg1& b ) : f(a), x(b) {}
 
@@ -114,15 +117,15 @@ struct Bind1of2 {
       return binder1of2<Binary,Arg1>(f,x);
    }
 };
-namespace {
-Bind1of2 bind1of2;
+FCPP_MAYBE_NAMESPACE_OPEN
+FCPP_MAYBE_EXTERN Bind1of2 bind1of2;
 //Curryable2<Bind1of2> bind1of2(_bind1of2);
-}
+FCPP_MAYBE_NAMESPACE_CLOSE
 
 template <class Binary, class Arg2>
 class binder2of2 {
-   Binary f;
-   Arg2 y;
+   const Binary f;
+   const Arg2 y;
 public:
    binder2of2( const Binary& a, const Arg2& b ) : f(a), y(b) {}
 
@@ -148,20 +151,17 @@ struct Bind2of2 {
       return binder2of2<Binary,Arg2>(f,y);
    }
 };
-namespace {
-Bind2of2 bind2of2;
+FCPP_MAYBE_NAMESPACE_OPEN
+FCPP_MAYBE_EXTERN Bind2of2 bind2of2;
 //Curryable2<Bind2of2> bind2of2(_bind2of2);
-}
+FCPP_MAYBE_NAMESPACE_CLOSE
 
 template <class Binary, class Arg1, class Arg2>
 class binder1and2of2 
-//: public CFunType<typename RT<Binary,Arg1,Arg2>::ResultType > {
-{
-   typedef typename RT<Binary,Arg1,Arg2>::ResultType ResultType;
-   struct Sig : public FunType<typename RT<Binary,Arg1,Arg2>::ResultType> {};
-   Binary f;
-   Arg1 a1;
-   Arg2 a2;
+: public CFunType<typename RT<Binary,Arg1,Arg2>::ResultType > {
+   const Binary f;
+   const Arg1 a1;
+   const Arg2 a2;
 public:
    binder1and2of2( const Binary& x, const Arg1& y, const Arg2& z ) 
    : f(x), a1(y), a2(z) {}
@@ -180,10 +180,10 @@ struct Bind1and2of2 {
       return binder1and2of2<Binary,Arg1,Arg2>(f,a1,a2);
    }
 };
-namespace {
-Bind1and2of2 bind1and2of2;
+FCPP_MAYBE_NAMESPACE_OPEN
+FCPP_MAYBE_EXTERN Bind1and2of2 bind1and2of2;
 //Curryable3<Bind1and2of2> bind1and2of2(_bind1and2of2);
-}
+FCPP_MAYBE_NAMESPACE_CLOSE
 
 //////////////////////////////////////////////////////////////////////
 // Now that bindNof2 are defined, we can define Curryable2, which then
@@ -213,7 +213,7 @@ struct Curryable2Helper<R,F,X,AutoCurryType> {
 
 template <class F>
 class Curryable2 {
-   F f;
+   const F f;
 public:
    Curryable2( const F& ff ) : f(ff) {}
 
@@ -259,10 +259,10 @@ Curryable2<F> makeCurryable2( const F& f ) {
 template <class Ternary, class A1, class A2, class A3>
 class binder1and2and3of3
 : public CFunType<typename RT<Ternary,A1,A2,A3>::ResultType> {
-   Ternary f;
-   A1 a1;
-   A2 a2;
-   A3 a3;
+   const Ternary f;
+   const A1 a1;
+   const A2 a2;
+   const A3 a3;
 public:
    binder1and2and3of3( const Ternary& w, const A1& x, const A2& y, const A3& z )
    : f(w), a1(x), a2(y), a3(z) {}
@@ -283,15 +283,15 @@ struct Bind1and2and3of3 {
       return binder1and2and3of3<Ternary,A1,A2,A3>(f,a1,a2,a3);
    }
 };
-namespace {
-Bind1and2and3of3 bind1and2and3of3; 
-}
+FCPP_MAYBE_NAMESPACE_OPEN
+FCPP_MAYBE_EXTERN Bind1and2and3of3 bind1and2and3of3; 
+FCPP_MAYBE_NAMESPACE_CLOSE
 
 template <class Ternary, class Arg1, class Arg2>
 class binder1and2of3 {
-   Ternary f;
-   Arg1 a1;
-   Arg2 a2;
+   const Ternary f;
+   const Arg1 a1;
+   const Arg2 a2;
 public:
    template <class Arg3>
    struct Sig 
@@ -316,16 +316,16 @@ struct Bind1and2of3 {
       return binder1and2of3<Ternary,A1,A2>(f,a1,a2);
    }
 };
-namespace {
-Bind1and2of3 bind1and2of3;
+FCPP_MAYBE_NAMESPACE_OPEN
+FCPP_MAYBE_EXTERN Bind1and2of3 bind1and2of3;
 //Curryable3<Bind1and2of3> bind1and2of3(_bind1and2of3);
-}
+FCPP_MAYBE_NAMESPACE_CLOSE
 
 template <class Ternary, class Arg2, class Arg3>
 class binder2and3of3 {
-   Ternary f;
-   Arg2 a2;
-   Arg3 a3;
+   const Ternary f;
+   const Arg2 a2;
+   const Arg3 a3;
 public:
    template <class Arg1>
    struct Sig 
@@ -352,16 +352,16 @@ struct Bind2and3of3 {
       return binder2and3of3<Ternary,A2,A3>(f,a2,a3);
    }
 };
-namespace {
-Bind2and3of3 bind2and3of3;
+FCPP_MAYBE_NAMESPACE_OPEN
+FCPP_MAYBE_EXTERN Bind2and3of3 bind2and3of3;
 //Curryable3<Bind2and3of3> bind2and3of3(_bind2and3of3);
-}
+FCPP_MAYBE_NAMESPACE_CLOSE
 
 template <class Ternary, class Arg1, class Arg3>
 class binder1and3of3 {
-   Ternary f;
-   Arg1 a1;
-   Arg3 a3;
+   const Ternary f;
+   const Arg1 a1;
+   const Arg3 a3;
 public:
    template <class Arg2>
    struct Sig 
@@ -388,15 +388,15 @@ struct Bind1and3of3 {
       return binder1and3of3<Ternary,A1,A3>(f,a1,a3);
    }
 };
-namespace {
-Bind1and3of3 bind1and3of3;
+FCPP_MAYBE_NAMESPACE_OPEN
+FCPP_MAYBE_EXTERN Bind1and3of3 bind1and3of3;
 //Curryable3<Bind1and3of3> bind1and3of3(_bind1and3of3);
-}
+FCPP_MAYBE_NAMESPACE_CLOSE
 
 template <class Ternary, class Arg1>
 class binder1of3 {
-   Ternary f;
-   Arg1 x;
+   const Ternary f;
+   const Arg1 x;
 public:
    binder1of3( const Ternary& a, const Arg1& b ) : f(a), x(b) {}
 
@@ -442,15 +442,15 @@ struct Bind1of3 {
       return makeCurryable2( binder1of3<Ternary,Arg1>(f,x) );
    }
 };
-namespace {
-Bind1of3 _bind1of3;
-Curryable2<Bind1of3> bind1of3(_bind1of3);
-}
+FCPP_MAYBE_NAMESPACE_OPEN
+FCPP_MAYBE_EXTERN Bind1of3 _bind1of3;
+FCPP_MAYBE_EXTERN Curryable2<Bind1of3> bind1of3 FCPP_MAYBE_INIT((_bind1of3));
+FCPP_MAYBE_NAMESPACE_CLOSE
 
 template <class Ternary, class Arg2>
 class binder2of3 {
-   Ternary f;
-   Arg2 x;
+   const Ternary f;
+   const Arg2 x;
 public:
    binder2of3( const Ternary& a, const Arg2& b ) : f(a), x(b) {}
 
@@ -478,15 +478,15 @@ struct Bind2of3 {
       return makeCurryable2( binder2of3<Ternary,Arg2>(f,x) );
    }
 };
-namespace {
-Bind2of3 _bind2of3;
-Curryable2<Bind2of3> bind2of3(_bind2of3);
-}
+FCPP_MAYBE_NAMESPACE_OPEN
+FCPP_MAYBE_EXTERN Bind2of3 _bind2of3;
+FCPP_MAYBE_EXTERN Curryable2<Bind2of3> bind2of3 FCPP_MAYBE_INIT((_bind2of3));
+FCPP_MAYBE_NAMESPACE_CLOSE
 
 template <class Ternary, class Arg3>
 class binder3of3 {
-   Ternary f;
-   Arg3 x;
+   const Ternary f;
+   const Arg3 x;
 public:
    binder3of3( const Ternary& a, const Arg3& b ) : f(a), x(b) {}
 
@@ -514,10 +514,10 @@ struct Bind3of3 {
       return makeCurryable2( binder3of3<Ternary,Arg3>(f,x) );
    }
 };
-namespace {
-Bind3of3 _bind3of3;
-Curryable2<Bind3of3> bind3of3(_bind3of3);
-}
+FCPP_MAYBE_NAMESPACE_OPEN
+FCPP_MAYBE_EXTERN Bind3of3 _bind3of3;
+FCPP_MAYBE_EXTERN Curryable2<Bind3of3> bind3of3 FCPP_MAYBE_INIT((_bind3of3));
+FCPP_MAYBE_NAMESPACE_CLOSE
 
 //////////////////////////////////////////////////////////////////////
 // "curry" versions. Now that we have auto-currying, you rarely need to
@@ -564,9 +564,9 @@ struct Curry3 {
     return binder1of3<Ternary,A1>(f,a1);
   }
 };
-namespace {
-Curry3 curry3; 
-}
+FCPP_MAYBE_NAMESPACE_OPEN
+FCPP_MAYBE_EXTERN Curry3 curry3; 
+FCPP_MAYBE_NAMESPACE_CLOSE
 
 struct Curry2 {
   template <class Binary, class A1, class A2 = void>
@@ -590,9 +590,9 @@ struct Curry2 {
     return binder1of2<Binary,A1>(f,a1);
   }
 };
-namespace {
-Curry2 curry2;
-}
+FCPP_MAYBE_NAMESPACE_OPEN
+FCPP_MAYBE_EXTERN Curry2 curry2;
+FCPP_MAYBE_NAMESPACE_CLOSE
 
 struct Curry1 {
   template <class Unary, class A1>
@@ -604,9 +604,9 @@ struct Curry1 {
     return binder1of1<Unary,A1>(f,a1);
   }
 };
-namespace {
-Curry1 curry1, curry; // "curry" is the same as "curry1"
-}
+FCPP_MAYBE_NAMESPACE_OPEN
+FCPP_MAYBE_EXTERN Curry1 curry1, curry; // "curry" is the same as "curry1"
+FCPP_MAYBE_NAMESPACE_CLOSE
 
 //////////////////////////////////////////////////////////////////////
 // Finally, Curryable3 (what a huge beast)
@@ -690,7 +690,7 @@ struct Curryable3Helper2<R,F,X,AutoCurryType> {
 
 template <class F>
 class Curryable3 {
-   F f;
+   const F f;
 public:
    Curryable3( const F& ff ) : f(ff) {}
 
