@@ -1,5 +1,5 @@
 //
-// Copyright (c) 2000,2001 Brian McNamara and Yannis Smaragdakis
+// Copyright (c) 2000,2001,2002 Brian McNamara and Yannis Smaragdakis
 //
 // Permission to use, copy, modify, distribute and sell this software
 // and its documentation for any purpose is granted without fee,
@@ -70,7 +70,10 @@ class Fun0 {
 public:
    // See comment in Fun0Impl about g++2.95.2
    typedef Result ResultType;
-   struct Sig { typedef Result ResultType; };
+   template <class Dummy1=Void, class Dummy2=Void>
+   struct Sig : public FunType<WrongNumberOfSigArgs> {};
+   template <class Dummy>
+   struct Sig<Void,Dummy> : public FunType<Result> {};
 
    typedef const Fun0Impl<Result>* Impl;
    // int is dummy arg to differentiate from the template constructor
@@ -80,7 +83,13 @@ public:
 
    template <class DF>   // direct functoid (or subtype polymorphism)
    Fun0( const DF& f ) : ref( Fun0Constructor<Result,DF>::make(f) ) {}
+
+   Fun0( const Fun0& x ) : ref(x.ref) {}
+   Fun0& operator=( const Fun0& x ) { ref = x.ref; return *this; }
 };
+// See comments below
+template <class R> struct Inherits<Fun0<R>,CallableWithoutArguments> 
+{ static const bool value = true; };
 
 template <class Result>
 struct Fun0Impl : public IRefable {
@@ -89,11 +98,18 @@ struct Fun0Impl : public IRefable {
    // CFunType class.  Fun0s are on the 'critical path' for lists, and need 
    // to be extra efficient.
    typedef Result ResultType;
-   struct Sig { typedef Result ResultType; };
+   template <class Dummy1=Void, class Dummy2=Void>
+   struct Sig : public FunType<WrongNumberOfSigArgs> {};
+   template <class Dummy>
+   struct Sig<Void,Dummy> : public FunType<Result> {};
 
    virtual Result operator()() const =0;
    virtual ~Fun0Impl() {}
 };
+// Since we cheated inheritance above, we need to inform our inheritance
+// detector for the particular case of importance.
+template <class R> struct Inherits<Fun0Impl<R>,CallableWithoutArguments> 
+{ static const bool value = true; };
 
 template <class Rd, class Rs>
 class Fun0Converter : public Fun0Impl<Rd> {
@@ -128,16 +144,16 @@ Fun0<Rd> explicit_convert0( const Fun0<Rs>& f ) {
 }
 
 template <class Gen>
-class Gen0 : public Fun0Impl<typename Gen::Sig::ResultType> {
+class Gen0 : public Fun0Impl<typename Gen::template Sig<>::ResultType> {
    Gen g;
 public:
    Gen0( Gen x ) : g(x) {}
-   typename Gen::Sig::ResultType operator()() const { return g(); }
+   typename Gen::template Sig<>::ResultType operator()() const { return g(); }
 };
    
-template <class Gen>
-Fun0<typename Gen::Sig::ResultType> makeFun0( const Gen& g ) {
-   return Fun0<typename Gen::Sig::ResultType>(1,new Gen0<Gen>(g));
+template <class F> 
+Fun0<typename RT<F>::ResultType> makeFun0( const F& g ) {
+   return Fun0<typename RT<F>::ResultType>(1,new Gen0<F>(g));
 }
 
 template <class Nullary>
@@ -195,6 +211,9 @@ public:
 
    template <class DF>
    Fun1( const DF& df ) : ref( Fun1Constructor<Arg1,Result,DF>::make(df) ) {}
+
+   Fun1( const Fun1& x ) : ref(x.ref) {}
+   Fun1& operator=( const Fun1& x ) { ref = x.ref; return *this; }
 };
 
 template <class Arg1, class Result>
@@ -327,6 +346,9 @@ public:
    Fun2( const DF& df ) 
    : ref( Fun2Constructor<Arg1,Arg2,Result,DF>::make(df) ) {}
 
+   Fun2( const Fun2& x ) : ref(x.ref) {}
+   Fun2& operator=( const Fun2& x ) { ref = x.ref; return *this; }
+
    // normal call
    Result operator()( const Arg1& x, const Arg2& y ) const { 
       return ref->operator()(x,y); 
@@ -478,6 +500,9 @@ public:
    template <class DF>
    Fun3Guts( const DF& df )
    : ref( Fun3Constructor<Arg1,Arg2,Arg3,Result,DF>::make(df) ) {}
+
+   Fun3Guts( const Fun3Guts& x ) : ref(x.ref) {}
+   Fun3Guts& operator=( const Fun3Guts& x ) { ref = x.ref; return *this; }
 };
 
 template <class Arg1, class Arg2, class Arg3, class Result>
@@ -498,6 +523,9 @@ public:
 
    template <class DF> 
    Fun3( const DF& df ) : rep(df) {}
+
+   Fun3( const Fun3& x ) : rep(x.rep) {}
+   Fun3& operator=( const Fun3& x ) { rep = x.rep; return *this; }
    
    typedef fcpp::Curryable3<Fun3Guts<Arg1,Arg2,Arg3,Result> > SigHelp;
    template <class A, class B=void, class C=void>
