@@ -1,5 +1,5 @@
 //
-// Copyright (c) 2000,2001 Brian McNamara and Yannis Smaragdakis
+// Copyright (c) 2000,2001,2002 Brian McNamara and Yannis Smaragdakis
 //
 // Permission to use, copy, modify, distribute and sell this software
 // and its documentation for any purpose is granted without fee,
@@ -25,6 +25,7 @@
 
 #include <functional>
 #include <utility>
+#include <iterator>
 #include "curry.h"
 
 namespace fcpp {
@@ -42,10 +43,10 @@ struct MakePair {
       return std::make_pair(a,b);
    }
 };
-namespace {
-MakePair _makePair;
-Curryable2<MakePair> makePair(_makePair);
-}
+FCPP_MAYBE_NAMESPACE_OPEN
+FCPP_MAYBE_EXTERN MakePair _makePair;
+FCPP_MAYBE_EXTERN Curryable2<MakePair> makePair FCPP_MAYBE_INIT((_makePair));
+FCPP_MAYBE_NAMESPACE_CLOSE
 
 struct Min {
    template <class T, class U> struct Sig;
@@ -58,10 +59,10 @@ struct Min {
       return std::less<T>()( x, y ) ? x : y;
    }
 };
-namespace {
-Min _min;
-Curryable2<Min> min(_min);
-}
+FCPP_MAYBE_NAMESPACE_OPEN
+FCPP_MAYBE_EXTERN Min _min;
+FCPP_MAYBE_EXTERN Curryable2<Min> min FCPP_MAYBE_INIT((_min));
+FCPP_MAYBE_NAMESPACE_CLOSE
 
 struct Max {
    template <class T, class U> struct Sig;
@@ -74,10 +75,10 @@ struct Max {
       return std::less<T>()( x, y ) ? y : x;
    }
 };
-namespace {
-Max _max;
-Curryable2<Max> max(_max);
-}
+FCPP_MAYBE_NAMESPACE_OPEN
+FCPP_MAYBE_EXTERN Max _max;
+FCPP_MAYBE_EXTERN Curryable2<Max> max FCPP_MAYBE_INIT((_max));
+FCPP_MAYBE_NAMESPACE_CLOSE
 
 struct Plus {
    template <class T, class U> struct Sig;
@@ -90,10 +91,10 @@ struct Plus {
       return std::plus<T>()( x, y );
    }
 };
-namespace {
-Plus _plus;
-Curryable2<Plus> plus(_plus);
-}
+FCPP_MAYBE_NAMESPACE_OPEN
+FCPP_MAYBE_EXTERN Plus _plus;
+FCPP_MAYBE_EXTERN Curryable2<Plus> plus FCPP_MAYBE_INIT((_plus));
+FCPP_MAYBE_NAMESPACE_CLOSE
 
 struct Minus {
    template <class T, class U> struct Sig;
@@ -106,10 +107,10 @@ struct Minus {
       return std::minus<T>()( x, y );
    }
 };
-namespace {
-Minus _minus;
-Curryable2<Minus> minus(_minus);
-}
+FCPP_MAYBE_NAMESPACE_OPEN
+FCPP_MAYBE_EXTERN Minus _minus;
+FCPP_MAYBE_EXTERN Curryable2<Minus> minus FCPP_MAYBE_INIT((_minus));
+FCPP_MAYBE_NAMESPACE_CLOSE
 
 struct Multiplies {
    template <class T, class U> struct Sig;
@@ -122,10 +123,10 @@ struct Multiplies {
       return std::multiplies<T>()( x, y );
    }
 };
-namespace {
-Multiplies _multiplies;
-Curryable2<Multiplies> multiplies(_multiplies);
-}
+FCPP_MAYBE_NAMESPACE_OPEN
+FCPP_MAYBE_EXTERN Multiplies _multiplies;
+FCPP_MAYBE_EXTERN Curryable2<Multiplies> multiplies FCPP_MAYBE_INIT((_multiplies));
+FCPP_MAYBE_NAMESPACE_CLOSE
 
 struct Divides {
    template <class T, class U> struct Sig;
@@ -138,10 +139,10 @@ struct Divides {
       return std::divides<T>()( x, y );
    }
 };
-namespace {
-Divides _divides;
-Curryable2<Divides> divides(_divides);
-}
+FCPP_MAYBE_NAMESPACE_OPEN
+FCPP_MAYBE_EXTERN Divides _divides;
+FCPP_MAYBE_EXTERN Curryable2<Divides> divides FCPP_MAYBE_INIT((_divides));
+FCPP_MAYBE_NAMESPACE_CLOSE
 
 struct Modulus {
    template <class T, class U> struct Sig;
@@ -154,10 +155,10 @@ struct Modulus {
       return std::modulus<T>()( x, y );
    }
 };
-namespace {
-Modulus _modulus;
-Curryable2<Modulus> modulus(_modulus);
-}
+FCPP_MAYBE_NAMESPACE_OPEN
+FCPP_MAYBE_EXTERN Modulus _modulus;
+FCPP_MAYBE_EXTERN Curryable2<Modulus> modulus FCPP_MAYBE_INIT((_modulus));
+FCPP_MAYBE_NAMESPACE_CLOSE
 
 struct Negate {
    template <class T>
@@ -168,9 +169,9 @@ struct Negate {
       return std::negate<T>()( x );
    }
 };
-namespace {
-Negate negate;
-}
+FCPP_MAYBE_NAMESPACE_OPEN
+FCPP_MAYBE_EXTERN Negate negate;
+FCPP_MAYBE_NAMESPACE_CLOSE
 
 struct Equal {
    template <class T, class U> struct Sig;
@@ -183,10 +184,10 @@ struct Equal {
       return std::equal_to<T>()( x, y );
    }
 };
-namespace {
-Equal _equal;
-Curryable2<Equal> equal(_equal);
-}
+FCPP_MAYBE_NAMESPACE_OPEN
+FCPP_MAYBE_EXTERN Equal _equal;
+FCPP_MAYBE_EXTERN Curryable2<Equal> equal FCPP_MAYBE_INIT((_equal));
+FCPP_MAYBE_NAMESPACE_CLOSE
 
 struct NotEqual {
    template <class T, class U> struct Sig;
@@ -199,10 +200,10 @@ struct NotEqual {
       return std::not_equal_to<T>()( x, y );
    }
 };
-namespace {
-NotEqual _notEqual;
-Curryable2<NotEqual> notEqual(_notEqual);
-}
+FCPP_MAYBE_NAMESPACE_OPEN
+FCPP_MAYBE_EXTERN NotEqual _notEqual;
+FCPP_MAYBE_EXTERN Curryable2<NotEqual> notEqual FCPP_MAYBE_INIT((_notEqual));
+FCPP_MAYBE_NAMESPACE_CLOSE
 
 struct Greater {
    template <class T, class U> struct Sig;
@@ -215,10 +216,10 @@ struct Greater {
       return std::greater<T>()( x, y );
    }
 };
-namespace {
-Greater _greater;
-Curryable2<Greater> greater(_greater);
-}
+FCPP_MAYBE_NAMESPACE_OPEN
+FCPP_MAYBE_EXTERN Greater _greater;
+FCPP_MAYBE_EXTERN Curryable2<Greater> greater FCPP_MAYBE_INIT((_greater));
+FCPP_MAYBE_NAMESPACE_CLOSE
 
 struct Less {
    template <class T, class U> struct Sig;
@@ -231,10 +232,10 @@ struct Less {
       return std::less<T>()( x, y );
    }
 };
-namespace {
-Less _less;
-Curryable2<Less> less(_less);
-}
+FCPP_MAYBE_NAMESPACE_OPEN
+FCPP_MAYBE_EXTERN Less _less;
+FCPP_MAYBE_EXTERN Curryable2<Less> less FCPP_MAYBE_INIT((_less));
+FCPP_MAYBE_NAMESPACE_CLOSE
 
 struct GreaterEqual {
    template <class T, class U> struct Sig;
@@ -247,10 +248,10 @@ struct GreaterEqual {
       return std::greater_equal<T>()( x, y );
    }
 };
-namespace {
-GreaterEqual _greaterEqual;
-Curryable2<GreaterEqual> greaterEqual(_greaterEqual);
-}
+FCPP_MAYBE_NAMESPACE_OPEN
+FCPP_MAYBE_EXTERN GreaterEqual _greaterEqual;
+FCPP_MAYBE_EXTERN Curryable2<GreaterEqual> greaterEqual FCPP_MAYBE_INIT((_greaterEqual));
+FCPP_MAYBE_NAMESPACE_CLOSE
 
 struct LessEqual {
    template <class T, class U> struct Sig;
@@ -263,10 +264,10 @@ struct LessEqual {
       return std::less_equal<T>()( x, y );
    }
 };
-namespace {
-LessEqual _lessEqual;
-Curryable2<LessEqual> lessEqual(_lessEqual);
-}
+FCPP_MAYBE_NAMESPACE_OPEN
+FCPP_MAYBE_EXTERN LessEqual _lessEqual;
+FCPP_MAYBE_EXTERN Curryable2<LessEqual> lessEqual FCPP_MAYBE_INIT((_lessEqual));
+FCPP_MAYBE_NAMESPACE_CLOSE
 
 struct LogicalAnd {
    template <class T, class U> struct Sig;
@@ -279,10 +280,10 @@ struct LogicalAnd {
       return std::logical_and<T>()( x, y );
    }
 };
-namespace {
-LogicalAnd _logicalAnd;
-Curryable2<LogicalAnd> logicalAnd(_logicalAnd);
-}
+FCPP_MAYBE_NAMESPACE_OPEN
+FCPP_MAYBE_EXTERN LogicalAnd _logicalAnd;
+FCPP_MAYBE_EXTERN Curryable2<LogicalAnd> logicalAnd FCPP_MAYBE_INIT((_logicalAnd));
+FCPP_MAYBE_NAMESPACE_CLOSE
 
 struct LogicalOr {
    template <class T, class U> struct Sig;
@@ -295,10 +296,10 @@ struct LogicalOr {
       return std::logical_or<T>()( x, y );
    }
 };
-namespace {
-LogicalOr _logicalOr;
-Curryable2<LogicalOr> logicalOr(_logicalOr);
-}
+FCPP_MAYBE_NAMESPACE_OPEN
+FCPP_MAYBE_EXTERN LogicalOr _logicalOr;
+FCPP_MAYBE_EXTERN Curryable2<LogicalOr> logicalOr FCPP_MAYBE_INIT((_logicalOr));
+FCPP_MAYBE_NAMESPACE_CLOSE
 
 struct LogicalNot {
    template <class T>
@@ -309,9 +310,35 @@ struct LogicalNot {
       return std::logical_not<T>()( x );
    }
 };
-namespace {
-LogicalNot logicalNot;
-}
+FCPP_MAYBE_NAMESPACE_OPEN
+FCPP_MAYBE_EXTERN LogicalNot logicalNot;
+FCPP_MAYBE_NAMESPACE_CLOSE
+
+struct Dereference {
+   template <class T> struct Sig 
+   : public FunType<T,typename std::iterator_traits<T>::value_type> {};
+
+   template <class T>
+   typename Sig<T>::ResultType operator()( const T& p ) const {
+      return *p;
+   }
+};
+FCPP_MAYBE_NAMESPACE_OPEN
+FCPP_MAYBE_EXTERN Dereference dereference;
+FCPP_MAYBE_NAMESPACE_CLOSE
+
+struct AddressOf {
+   template <class T>
+   struct Sig : public FunType<T,const T*> {};
+
+   template <class T>
+   const T* operator()( const T& x ) const {
+      return &x;
+   }
+};
+FCPP_MAYBE_NAMESPACE_OPEN
+FCPP_MAYBE_EXTERN AddressOf addressOf;
+FCPP_MAYBE_NAMESPACE_CLOSE
 
 // The other "higher-order" operators in STL are easy to do in our
 // framework. For instance, instead of unary_negate (with all its
@@ -323,7 +350,7 @@ LogicalNot logicalNot;
 //////////////////////////////////////////////////////////////////////
 
 template <class Op>
-class stl1 : CFunType<typename Op::argument_type,
+class stl1 : public CFunType<typename Op::argument_type,
                        typename Op::result_type> {
    Op f;
 public:
@@ -340,7 +367,7 @@ stl1<Op> stl_to_fun1( const Op& o ) {
 }
 
 template <class Op>
-class stl2 : CFunType<typename Op::first_argument_type,
+class stl2 : public CFunType<typename Op::first_argument_type,
                        typename Op::second_argument_type,
                        typename Op::result_type> {
    Op f;
@@ -507,6 +534,24 @@ ptr_to_fun(Result (Arg1::*x)(Arg2,Arg3)) {
             ptr_to_mem_binary_function<Arg1, Arg2, Arg3, Result>(x) );
 }
 
+template <class Arg1, class Arg2, class Arg3, class Result>
+class ptr_to_const_mem_binary_function 
+: public CFunType<Arg1*, Arg2, Arg3, Result> {
+    Result (Arg1::*ptr)(Arg2,Arg3) const;
+public:
+    explicit ptr_to_const_mem_binary_function(
+          Result (Arg1::*x)(Arg2,Arg3) const) : ptr(x) {}
+    Result operator()(const Arg1* x, const Arg2& y, const Arg3& z) const 
+       { return (x->*ptr)(y,z); }
+};
+
+template <class Arg1, class Arg2, class Arg3, class Result>
+inline Curryable3<ptr_to_const_mem_binary_function<Arg1, Arg2, Arg3, Result> >
+ptr_to_fun(Result (Arg1::*x)(Arg2,Arg3) const) {
+  return makeCurryable3(
+            ptr_to_const_mem_binary_function<Arg1, Arg2, Arg3, Result>(x) );
+}
+
 template <class Arg1, class Arg2, class Result>
 class ptr_to_mem_unary_function : public CFunType<Arg1*, Arg2, Result> {
     Result (Arg1::*ptr)(Arg2);
@@ -521,6 +566,23 @@ ptr_to_fun(Result (Arg1::*x)(Arg2)) {
   return makeCurryable2( ptr_to_mem_unary_function<Arg1, Arg2, Result>(x) );
 }
 
+template <class Arg1, class Arg2, class Result>
+class ptr_to_const_mem_unary_function : public CFunType<Arg1*, Arg2, Result> {
+    Result (Arg1::*ptr)(Arg2) const;
+public:
+    explicit ptr_to_const_mem_unary_function(Result (Arg1::*x)(Arg2) const) 
+       : ptr(x) {}
+    Result operator()(const Arg1* x, const Arg2& y) const 
+       { return (x->*ptr)(y); }
+};
+
+template <class Arg1, class Arg2, class Result>
+inline Curryable2<ptr_to_const_mem_unary_function<Arg1, Arg2, Result> >
+ptr_to_fun(Result (Arg1::*x)(Arg2) const) {
+  return makeCurryable2( 
+    ptr_to_const_mem_unary_function<Arg1, Arg2, Result>(x) );
+}
+
 template <class Arg1, class Result>
 class ptr_to_mem_nullary_function : public CFunType<Arg1*, Result> {
     Result (Arg1::*ptr)();
@@ -535,6 +597,21 @@ ptr_to_fun(Result (Arg1::*x)()) {
   return ptr_to_mem_nullary_function<Arg1, Result>(x);
 }
 
+template <class Arg1, class Result>
+class ptr_to_const_mem_nullary_function : public CFunType<Arg1*, Result> {
+    Result (Arg1::*ptr)() const;
+public:
+    explicit ptr_to_const_mem_nullary_function(Result (Arg1::*x)() const) 
+       : ptr(x) {}
+    Result operator()(const Arg1* x) const { return (x->*ptr)(); }
+};
+
+template <class Arg1, class Result>
+inline ptr_to_const_mem_nullary_function<Arg1, Result>
+ptr_to_fun(Result (Arg1::*x)() const) {
+  return ptr_to_const_mem_nullary_function<Arg1, Result>(x);
+}
+
 //////////////////////////////////////////////////////////////////////
 // Misc stuff
 //////////////////////////////////////////////////////////////////////
@@ -546,9 +623,9 @@ struct Inc {
     template <class T>
     T operator()(const T& x) const { T y = x; return ++y; }
 };
-namespace {
-Inc inc; 
-}
+FCPP_MAYBE_NAMESPACE_OPEN
+FCPP_MAYBE_EXTERN Inc inc; 
+FCPP_MAYBE_NAMESPACE_CLOSE
 
 struct Always1 {
     template <class T>
@@ -557,9 +634,9 @@ struct Always1 {
     template <class T>
     bool operator()(const T&) const { return true; }
 };
-namespace {
-Always1 always1;
-}
+FCPP_MAYBE_NAMESPACE_OPEN
+FCPP_MAYBE_EXTERN Always1 always1;
+FCPP_MAYBE_NAMESPACE_CLOSE
 
 struct Never1 {
     template <class T>
@@ -568,9 +645,9 @@ struct Never1 {
     template <class T>
     bool operator()(const T&) const { return false; }
 };
-namespace {
-Never1 never1;
-}
+FCPP_MAYBE_NAMESPACE_OPEN
+FCPP_MAYBE_EXTERN Never1 never1;
+FCPP_MAYBE_NAMESPACE_CLOSE
 
 } // end namespace fcpp
 
